import bs4
import re
from pathlib import Path
import mdformat
from collections.abc import Iterator

JAVADOC_URL_PREFIX = 'https://secata.gitlab.io'

DOCKER_USAGE = '''
## Usage

To build a local docker image, run:

```sh
mvn package com.google.cloud.tools:jib-maven-plugin:dockerBuild -pl {ARTIFACT_ID} -am
```

Then to run the docker image:

```sh
sudo docker run -v CONFIG_PATH_HERE":/conf/server.json" {ARTIFACT_ID}:latest
```

See <conf> directory for example configurations.
'''



def html_to_markdown(markdown_out: list[str], soup) -> None:
    assert soup is not None, soup
    if isinstance(soup, list):
        for e in soup:
            html_to_markdown(markdown_out, e)
        return

    if soup.name is None:
        markdown_out.append(soup)
    elif soup.name in {'p','section','div'}:
        markdown_out.append('\n\n')
        html_to_markdown(markdown_out, soup.contents)
        markdown_out.append('\n\n')
    elif soup.name == 'code':
        markdown_out.append('`')
        html_to_markdown(markdown_out, soup.contents)
        markdown_out.append('`')
    elif soup.name == 'a':
        markdown_out.append('[')
        html_to_markdown(markdown_out, soup.contents)
        markdown_out.append('](')
        markdown_out.append(soup['href'])
        markdown_out.append(')')
    elif soup.name == 'b':
        markdown_out.append('**')
        html_to_markdown(markdown_out, soup.contents)
        markdown_out.append('**')
    elif soup.name == 'i':
        markdown_out.append('_')
        html_to_markdown(markdown_out, soup.contents)
        markdown_out.append('_')
    elif soup.name == 'ul':
        html_to_markdown(markdown_out, soup.contents)
    elif soup.name == 'li':
        markdown_out.append('- ')
        html_to_markdown(markdown_out, soup.contents)
    elif soup.name == 'h2':
        markdown_out.append('## ')
        html_to_markdown(markdown_out, soup.contents)
    else:
        assert False, soup

def concat_markdown(parts: list[str]) -> str:
    new_readme = ''.join(parts).strip()
    new_readme = re.sub(r'\n\s*\n', '\n\n', new_readme)
    new_readme = re.sub(r'\n *(\S)', r'\n\1', new_readme)
    return mdformat.text(new_readme, options = { "wrap": 100 })

def add_further_links(markdown_out: list[str], javadoc_url: str) -> None:
    markdown_out.append('## Further links\n')
    markdown_out.append(f'- [Javadoc Documentation for `main` branch]({javadoc_url})\n')

def parse_pom(project_path: Path):
    with open(project_path / 'pom.xml') as f:
        return bs4.BeautifulSoup(f, 'lxml-xml')

DEPENDENCY_FORMAT = '''
<dependency>
    <groupId>{GROUP_ID}</groupId>
    <artifactId>{ARTIFACT_ID}</artifactId>
    <version>{VERSION}</version>
</dependency>
'''.strip()

def get_maven_group_artifact_id(pom) -> tuple[str,str]:
    group_id = pom.find('groupId').get_text().strip()
    artifact_id = pom.find('artifactId').get_text().strip()
    return (group_id, artifact_id)


def usage_maven_import_library(markdown_out: list[str], project_path: Path):
    (group_id, artifact_id) = get_maven_group_artifact_id(parse_pom(project_path))

    markdown_out.append('## Usage, Maven\n')
    markdown_out.append('In your Maven `pom.xml`, add following to the `<dependencies>` section:\n')
    markdown_out.append('```xml\n')
    markdown_out.append(DEPENDENCY_FORMAT.format(GROUP_ID=group_id, ARTIFACT_ID=artifact_id, VERSION='VERSION'))
    markdown_out.append('\n```\n\n')
    markdown_out.append('A suitable `VERSION` can be found in the package registry.\n')


def readme_for_repo_without_main(markdown_out: list[str]):
    markdown_out.append('Java Library.\n\n')

def readme_from_class(project_path: Path, class_path: str | None, extra_sections: list[str] = []) -> str:
    soup = None

    if class_path is not None:
        docs_file = (project_path/ 'target'/'reports'/'apidocs'/class_path.replace('.','/')).with_suffix('.html')
        with open(docs_file) as f:
            soup = bs4.BeautifulSoup(f, 'lxml')

    title = project_path.name.replace('-', ' ').capitalize()

    gitlab_project_path = project_path.relative_to('/home/jmaa/src/secata')

    javadoc_url = JAVADOC_URL_PREFIX + '/' + str(gitlab_project_path)

    markdown_out = ['# ', title,'\n\n']
    if soup is not None:
        html_to_markdown(markdown_out, soup.select_one('#class-description .block'))
    else:
        readme_for_repo_without_main(markdown_out )
        usage_maven_import_library(markdown_out, project_path)

    markdown_out += extra_sections
    add_further_links(markdown_out, javadoc_url)
    return concat_markdown(markdown_out)


def auto_readme_for_contract_maven_java(project_path: Path) -> str | None:
    main_resource_path = project_path / 'src'/'main'/'resources'/'main'
    if not main_resource_path.exists():
        return None

    with open(main_resource_path) as f:
        main_class_path = f.read().strip()

    return readme_from_class(project_path, main_class_path)


def auto_readme_for_docker_maven_java(project_path: Path) -> str | None:
    pom = parse_pom(project_path)
    class_soup = pom.find('configurator.class')
    (group_id, artifact_id) = get_maven_group_artifact_id(pom)
    if class_soup is None:
        return None

    main_class_path = class_soup.get_text().strip()

    extra_sections = [DOCKER_USAGE.format(ARTIFACT_ID=artifact_id)]

    return readme_from_class(project_path, main_class_path, extra_sections)


def auto_readme_for_generic_maven_java(project_path: Path) -> str:
    return readme_from_class(project_path, None, [])


def multi_maven_modules(project_path: Path) -> Iterator[str]:
    pom = parse_pom(project_path)

    for module_soup in pom.select('modules module'):
        yield module_soup.get_text()


def auto_readme_for_partisia_maven_java_multi_mavens(project_path: Path) -> None:
    title = project_path.name.replace('-', ' ').capitalize()

    markdown_out = ['# ', title, '\n\n']
    markdown_out.append('## Repository layout\n')
    markdown_out.append('Submodules:\n')

    for submodule in multi_maven_modules(project_path):
        markdown_out.append(f'- [`{submodule}`](./{submodule})\n')
    return concat_markdown(markdown_out)

def auto_readme_for_project(project_path: Path, generate_multi_maven: bool) -> None:
    new_readme = None

    # Check whether multi-module
    submodules = list(multi_maven_modules(project_path))
    if submodules:
        for submodule in submodules:
            auto_readme_for_project(project_path / submodule)
            del submodule
        if generate_multi_maven:
            new_readme = auto_readme_for_partisia_maven_java_multi_mavens(project_path)
        return

    # Assume contract
    if new_readme is None:
        new_readme = auto_readme_for_contract_maven_java(project_path)

    # Assume docker
    if new_readme is None:
        new_readme = auto_readme_for_docker_maven_java(project_path)

    # Assume general maven java
    if new_readme is None:
        new_readme = auto_readme_for_generic_maven_java(project_path)

    assert new_readme is not None, project_path

    readme_path = project_path / 'README.md'
    with open(readme_path, 'w') as f:
        f.write(new_readme)


